from channels.routing import ProtocolTypeRouter, URLRouter
from channels.auth import AuthMiddlewareStack
import core.routing

application = ProtocolTypeRouter(
    {"websocket": AuthMiddlewareStack(URLRouter(core.routing.websocket_urlpatterns)),}
)
